/*
 *  jDTAUS - DTAUS fileformat.
 *  Copyright (c) 2005 Christian Schulte <cs@schulte.it>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
package org.jdtaus.core.monitor.ri.server;

import java.util.Date;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import javax.swing.event.EventListenerList;
import org.jdtaus.core.container.ContainerFactory;
import org.jdtaus.core.container.ContainerInitializer;
import org.jdtaus.core.container.ContextFactory;
import org.jdtaus.core.container.ContextInitializer;
import org.jdtaus.core.container.Dependency;
import org.jdtaus.core.container.Implementation;
import org.jdtaus.core.container.ModelFactory;
import org.jdtaus.core.container.Properties;
import org.jdtaus.core.container.Property;
import org.jdtaus.core.container.PropertyException;
import org.jdtaus.core.monitor.Task;
import org.jdtaus.core.monitor.TaskEvent;
import org.jdtaus.core.monitor.TaskListener;
import org.jdtaus.core.logging.spi.Logger;
import org.jdtaus.core.monitor.spi.TaskMonitor;

/**
 * jDTAUS Core SPI Server {@code TaskMonitor} implementation.
 * <p>The server implementation will measure the time a task is running. The
 * configuration property {@code longRunningMillis} holds the amount of
 * milliseconds a task needs to run before it is considered a long running task.
 * The server implementation will then log notification messages about long
 * running tasks. Since this implementation does not monitor a task for changes
 * {@code TaskEvent}s of type {@code CHANGED_STATE} may only be fired right
 * before the last event of type {@code ENDED} is fired.</p>
 *
 * @author <a href="mailto:cs@schulte.it">Christian Schulte</a>
 * @version $Id: ServerTaskMonitor.java 3553 2007-05-25 22:37:09Z schulte2005 $
 *
 * @see org.jdtaus.core.container.Container
 * @see #initialize()
 */
public final class ServerTaskMonitor
    implements TaskMonitor, ContainerInitializer
{
    //--Implementation----------------------------------------------------------

    // This section is managed by jdtaus-container-mojo.

    /** Meta-data describing the implementation. */
    private static final Implementation META =
        ModelFactory.getModel().getModules().
        getImplementation(ServerTaskMonitor.class.getName());

    //----------------------------------------------------------Implementation--
    //--Constructors------------------------------------------------------------

    // This section is managed by jdtaus-container-mojo.

    /**
     * Protected <code>ServerTaskMonitor</code> implementation constructor.
     *
     * @param meta Implementation meta-data.
     *
     * @throws NullPointerException if <code>meta</code> is <code>null</code>.
     */
    protected ServerTaskMonitor(final Implementation meta)
    {
        super();
        if(meta == null)
        {
            throw new NullPointerException("meta");
        }
        this.initializeProperties(meta.getProperties());
    }
    /**
     * Protected <code>ServerTaskMonitor</code> dependency constructor.
     *
     * @param meta dependency meta-data.
     *
     * @throws NullPointerException if <code>meta</code> is <code>null</code>.
     */
    protected ServerTaskMonitor(final Dependency meta)
    {
        super();
        if(meta == null)
        {
            throw new NullPointerException("meta");
        }
        this.initializeProperties(meta.getProperties());
    }

    /**
     * Initializes the properties of the instance.
     *
     * @param meta the property values to initialize the instance with.
     *
     * @throws NullPointerException if {@code meta} is {@code null}.
     */
    protected void initializeProperties(final Properties meta)
    {
        Property p;

        if(meta == null)
        {
            throw new NullPointerException("meta");
        }

        p = meta.getProperty("longRunningMillis");
        this._longRunningMillis = ((java.lang.Long) p.getValue()).longValue();

    }

    //------------------------------------------------------------Constructors--
    //--ContainerInitializer----------------------------------------------------

    /**
     * Initializes the instance.
     *
     * @see #assertValidProperties()
     */
    public void initialize()
    {
        this.assertValidProperties();
    }

    //----------------------------------------------------ContainerInitializer--
    //--Properties--------------------------------------------------------------

    // This section is managed by jdtaus-container-mojo.

    /**
     * Property {@code longRunningMillis}.
     * @serial
     */
    private long _longRunningMillis;

    /**
     * Gets the value of property <code>longRunningMillis</code>.
     *
     * @return the value of property <code>longRunningMillis</code>.
     */
    protected long getLongRunningMillis()
    {
        return this._longRunningMillis;
    }


    //--------------------------------------------------------------Properties--
    //--Dependencies------------------------------------------------------------

    // This section is managed by jdtaus-container-mojo.

    /** Configured <code>Logger</code> implementation. */
    private transient Logger _dependency0;

    /**
     * Gets the configured <code>Logger</code> implementation.
     *
     * @return the configured <code>Logger</code> implementation.
     */
    private Logger getLogger()
    {
        Logger ret = null;
        if(this._dependency0 != null)
        {
            ret = this._dependency0;
        }
        else
        {
            ret = (Logger) ContainerFactory.getContainer().
                getDependency(ServerTaskMonitor.class,
                "Logger");

            if(ModelFactory.getModel().getModules().
                getImplementation(ServerTaskMonitor.class.getName()).
                getDependencies().getDependency("Logger").
                isBound())
            {
                this._dependency0 = ret;
            }
        }

        if(ret instanceof ContextInitializer && !((ContextInitializer) ret).
            isInitialized(ContextFactory.getContext()))
        {
            ((ContextInitializer) ret).initialize(ContextFactory.getContext());
        }

        return ret;
    }

    //------------------------------------------------------------Dependencies--
    //--TaskEventSource---------------------------------------------------------

    /** List of {@code TaskListener}s. */
    private final EventListenerList taskListeners = new EventListenerList();

    public void addTaskListener(final TaskListener listener)
    {
        if(listener == null)
        {
            throw new NullPointerException("listener");
        }

        this.taskListeners.add(TaskListener.class, listener);
    }

    public void removeTaskListener(final TaskListener listener)
    {
        if(listener == null)
        {
            throw new NullPointerException("listener");
        }

        this.taskListeners.remove(TaskListener.class, listener);
    }

    public TaskListener[] getTaskListeners()
    {
        return (TaskListener[]) this.taskListeners.
            getListeners(TaskListener.class);

    }

    //---------------------------------------------------------TaskEventSource--
    //--TaskMonitor-------------------------------------------------------------

    public void monitor(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        synchronized(this.mutex)
        {
            this.createTaskState(task);
            this.fireTaskEvent(new TaskEvent(task, TaskEvent.STARTED));
        }
    }

    public void finish(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        synchronized(this.mutex)
        {
            final long start = this.getStartTime(task);
            if(System.currentTimeMillis() - start > this.getLongRunningMillis())
            {
                this.getLogger().warn(ServerTaskMonitorBundle.
                    getLongRunningTaskNotificationMessage(
                    Locale.getDefault()).format( new Object[] {
                    task.getDescription().getText(Locale.getDefault()),
                    new Date(start), new Date(System.currentTimeMillis())
                }));
            }

            if(this.changedState(task))
            {
                this.fireTaskEvent(new TaskEvent(task,
                    TaskEvent.CHANGED_STATE));

            }

            this.removeTaskState(task);
            this.fireTaskEvent(new TaskEvent(task, TaskEvent.ENDED));
        }
    }

    //-------------------------------------------------------------TaskMonitor--
    //--ServerTaskMonitor-------------------------------------------------------

    /** Caches the state of a task. */
    private static class TaskState
    {
        private boolean indeterminate;
        private boolean cancelable;
        private boolean cancelled;
        private int minimum;
        private int maximum;
        private int progress;
    }

    /** Maps {@code Task}s to theire starting timestamp. */
    private final Map startMap = new HashMap(100);

    /** Maps {@code Task}s to a {@code TaskState} instance. */
    private final Map stateMap = new HashMap(100);

    /** Mutex to synchronize access to the maps. */
    private final Object mutex = new Object();

    /** Creates a new {@code ServerTaskMonitor} instance. */
    public ServerTaskMonitor()
    {
        this(ServerTaskMonitor.META);
        this.initialize();
    }

    /**
     * Checks configured properties.
     *
     * @throws PropertyException for illegal property values.
     */
    protected void assertValidProperties()
    {
        if(this.getLongRunningMillis() < 0L)
        {
            throw new PropertyException("longRunningMillies",
                Long.toString(this.getLongRunningMillis()));

        }
    }

    /**
     * Notifies all registered {@code TaskListener}s about {@code TaskEvent}s.
     *
     * @param e The event to be provided to the listeners.
     */
    protected void fireTaskEvent(final TaskEvent e)
    {
        if(e == null)
        {
            throw new NullPointerException("e");
        }

        final Object[] listeners = this.taskListeners.getListenerList();
        for (int i = listeners.length - 2; i >= 0; i -= 2)
        {
            if (listeners[i] == TaskListener.class)
            {
                ((TaskListener) listeners[i + 1]).onTaskEvent(e);
            }
        }
    }

    /**
     * Caches the state of a {@code Task}.
     *
     * @param task the task to cache state for.
     *
     * @throws NullPointerException if {@code task} is {@code null}.
     * @throws IllegalStateException if the cache already holds state for
     * {@code task}.
     */
    private void createTaskState(final Task task)
    {
        final TaskState state = new TaskState();
        state.cancelable = task.isCancelable();
        state.indeterminate = task.isIndeterminate();
        state.cancelled = state.cancelable ? task.isCancelled() : false;
        if(state.indeterminate)
        {
            state.maximum = Integer.MIN_VALUE;
            state.minimum = Integer.MIN_VALUE;
            state.progress = Integer.MIN_VALUE;
        }
        else
        {
            state.maximum = task.getMaximum();
            state.minimum = task.getMinimum();
            state.progress = task.getProgress();
        }

        if(this.stateMap.put(task, state) != null || this.startMap.put(
            task, new Long(System.currentTimeMillis())) != null)
        {
            throw new IllegalStateException();
        }
    }

    /**
     * Removes the cached state of a {@code Task}.
     *
     * @param task the task to remove the cached state of.
     *
     * @throws NullPointerException if {@code task} is {@code null}.
     */
    private void removeTaskState(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        this.stateMap.remove(task);
        this.startMap.remove(task);
    }

    /**
     * Checks the state of a given task for changes.
     *
     * @param task the task to check for state changes.
     *
     * @return {@code true} if the state of {@code task} changed since the last
     * time this method got called; {@code false} if the state did not change.
     *
     * @throws NullPointerException if {@code task} is {@code null}.
     * @throws IllegalStateException if no cached state exists for {@code task}.
     */
    private boolean changedState(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        boolean changedState = false;
        final TaskState state = (TaskState) this.stateMap.get(task);

        if(state == null)
        {
            throw new IllegalStateException();
        }

        if(state.indeterminate)
        {
            state.indeterminate = task.isIndeterminate();
            if(!state.indeterminate)
            {
                state.minimum = task.getMinimum();
                state.maximum = task.getMaximum();
                state.progress = task.getProgress();
                changedState = true;
            }
        }
        else
        {
            state.indeterminate = task.isIndeterminate();
            if(state.indeterminate)
            {
                changedState = true;
            }
            else
            {
                if(state.minimum != task.getMinimum())
                {
                    state.minimum = task.getMinimum();
                    changedState = true;
                }
                if(state.maximum != task.getMaximum())
                {
                    state.maximum = task.getMaximum();
                    changedState = true;
                }
                if(state.progress != task.getProgress())
                {
                    state.progress = task.getProgress();
                    changedState = true;
                }
            }
        }

        if(state.cancelable)
        {
            state.cancelable = task.isCancelable();
            if(!state.cancelable)
            {
                changedState = true;
            }
            else
            {
                if(state.cancelled != task.isCancelled())
                {
                    state.cancelled = task.isCancelled();
                    changedState = true;
                }
            }
        }
        else
        {
            state.cancelable = task.isCancelable();
            if(!state.cancelable)
            {
                state.cancelled = false;
                changedState = true;
            }
        }

        return changedState;
    }

    /**
     * Gets the timestamp a given task started.
     *
     * @param task the task to return the starting timestamp for.
     *
     * @throws NullPointerException if {@code task} is {@code null}.
     * @throws IllegalStateException if no timestamp information was cached for
     * {@code task}.
     */
    private long getStartTime(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        final Long l = (Long) this.startMap.get(task);
        if(l == null)
        {
            throw new IllegalStateException();
        }

        return l.longValue();
    }

    //-------------------------------------------------------ServerTaskMonitor--
}
